@@ -18,10 +18,13 @@ import tempfile
 import os
 import warnings
 import time
+from distutils.version import LooseVersion, StrictVersion
 
-from .protocol import connection_pb2
+from .protocol import connection_pb2, error_pb2
 from .protocol.stream import connection_pb2 as stream_connection_pb2
 from .protocol import options_pb2, common_pb2
+from .protocol.config import algorithm_pb2
+from .protocol.status import time_synchronization_pb2
 from . import stream
 from .version import __version__
 
@@ -35,6 +38,8 @@ class scanner(object):
     :type hostname_or_ip: str
     :param port: Port on which the device is reachable the default is 8000.
     :type port: int
+    :param timeout: Timeout in seconds until connection is aborted
+    :type timeout: int
     :param name: Name of the device used for string representation.
     :type name: str
     :param key_and_cert_file: Filename containing a private key and certificate for SSL connection.
@@ -46,8 +51,9 @@ class scanner(object):
     _default_port = 8000
     _default_ssl_port = 8800
     
-    def __init__(self, hostname_or_ip="localhost", port=None, name=None, key_and_cert_file=None, key_and_cert=None):
+    def __init__(self, hostname_or_ip="localhost", port=None, timeout=30, name=None, key_and_cert_file=None, key_and_cert=None):
         self.hostname_or_ip = hostname_or_ip
+        self.timeout = timeout
         self.name = name if name is not None else self.hostname_or_ip
         self.api_path = "http://" + hostname_or_ip + "/api/v1/"
         
@@ -66,9 +72,19 @@ class scanner(object):
         
         self._connection = self.create_connection()
         
-        hello = self.hello()
-        if __version__ != hello.library_version:
-            warnings.warn("Warning: The client BSL version does not match the server BSL version. Client has '%s', server has '%s'." % (__version__, hello.library_version), DeprecationWarning, stacklevel=2)
+        self._hello = self.hello()
+        # Compare library version of server and client
+        if __version__ != self._hello.library_version:
+            def parse_version(version_str):
+                return [int(x) for x in version_str.split(".")]
+            
+            local_version = None
+            if self._hello.library_version and __version__:
+                local_version = [int(x) if x.isdecimal() else None for x in __version__.split(".")]
+                server_version = [int(x) if x.isdecimal() else None for x in self._hello.library_version.split(".")]
+            # Only warn if major or minor version does not match
+            if local_version is None or len(local_version) < 3 or len(server_version) < 3 or local_version[0] != server_version[0] or local_version[1] != server_version[1]:
+                warnings.warn("Warning: The client BSL version does not match the server BSL version. Client has '%s', server has '%s'." % (__version__, self._hello.library_version), DeprecationWarning, stacklevel=2)
 
     def __del__(self):
         if hasattr(self, "_connection"):
@@ -128,7 +144,7 @@ class scanner(object):
         """
         return stream.status(self.create_connection())
 
-    def get_point_cloud_stream(self, filter=None, reference_frame=None, point_filter=None, as_numpy=False):
+    def get_point_cloud_stream(self, filter=None, reference_frame=None, point_filter=None, algorithms=None, as_numpy=False):
         """ Request point cloud stream of device
 
         :param filter: DEPRECATED
@@ -144,6 +160,10 @@ class scanner(object):
             > Introduced in BSL v2.13 and firmware v1.9
 
             Filter points and returns by point attributes during the post-processing on the device. This replaces the 'filter' parameter
+        :param algorithms:
+            > Introduced in BSL v2.17 and firmware v1.16
+
+            Add set of algorithms which post-process the point cloud data.
         :param as_numpy:
             > Introduced in BSL v2.16
     
@@ -159,7 +179,7 @@ class scanner(object):
         if filter:
             warnings.warn("Please use 'point_filter' instead of 'filter'. 'filter' is deprecated.", DeprecationWarning, stacklevel=2)
             point_filter = filter
-        return stream.point_cloud(self.create_connection(), point_filter=point_filter, reference_frame=reference_frame, as_numpy=as_numpy)
+        return stream.point_cloud(self.create_connection(), point_filter=point_filter, reference_frame=reference_frame, algorithms=algorithms, as_numpy=as_numpy)
 
     def get_raw_point_cloud_stream(self, point_filter=None, reference_frame=None):
         """ Request raw point cloud stream of device
@@ -210,16 +230,65 @@ class scanner(object):
             req.point_cloud.reference_frame.CopyFrom(reference_frame)
 
         return stream.raw(self.create_connection(), req, file_name)
-
+        
     @staticmethod
-    def file_point_cloud_stream(dump_filename):
+    def file_point_cloud_stream(dump_filename, as_numpy=False):
         """ Request a point_cloud_stream, which streams off a .bfpc file.
         No device (and connection to a device) is needed for this operation.
 
         :param dump_filename: path to .bfpc file
+        :param as_numpy:
+            > Introduced in BSL v2.16
+
+            This enables numpy support of the point cloud stream. Use recv_frame_as_numpy() to fetch frames as numpy structured arrays.
+            Specify the required attributes via the reference_frame attribute.
+
+            Note: The performance of this option is significantly better with files that were recorded in the packed format. Packed recordings are available in firmware versions >= v1.17.
         :return: :py:class:`blickfeld_scanner.scanner.stream.point_cloud` object
         """
-        return stream.point_cloud(from_file=dump_filename)
+        return stream.point_cloud(from_file=dump_filename, as_numpy=as_numpy)
+
+    def set_default_point_cloud_algorithms(self, persist=False, background_subtraction=False, neighbor_filter=False, static_transformation=False):
+        """ Sets point cloud algorithms, which are then applied on all point cloud stream subscribe requests of clients.
+        Clients can still actively request other algorithms or disabled them.
+
+        Args:
+            persist (bool, optional): Persist the configuration on the device and activates it after a power-cycle. Defaults to False.
+            background_subtraction (bool, optional): Either enable the background subtraction with default parameters or supply a configuration object. Defaults to False.
+            neighbor_filter (bool, optional): Either enable the neighbor filter with default parameters or supply a configuration object Defaults to False.
+            static_transformation (bool, optional): Either enable the static transformation with default parameters or supply a configuration object. Defaults to False.
+
+        Returns:
+            Currently set advanced config, see :any:`protobuf_protocol` advanced config
+        """
+        advanced = self.get_advanced_config()
+        del advanced.server.default_point_cloud_subscription.algorithms[:]
+        
+        algorithms = []
+        if background_subtraction:
+            cfg = algorithm_pb2.Algorithm()
+            if type(background_subtraction) == bool:
+                cfg.background_subtraction.SetInParent()
+            else:
+                cfg.background_subtraction.CopyFrom(background_subtraction)
+            algorithms.append(cfg)
+        if neighbor_filter:
+            cfg = algorithm_pb2.Algorithm()
+            if type(neighbor_filter) == bool:
+                cfg.neighbor_filter.SetInParent()
+            else:
+                cfg.neighbor_filter.CopyFrom(neighbor_filter)
+            algorithms.append(cfg)
+        if static_transformation:
+            cfg = algorithm_pb2.Algorithm()
+            if type(static_transformation) == bool:
+                cfg.static_transformation.SetInParent()
+            else:
+                cfg.static_transformation.CopyFrom(static_transformation)
+            algorithms.append(cfg)
+        
+        advanced.server.default_point_cloud_subscription.algorithms.extend(algorithms)
+        return self.set_advanced_config(advanced, persist=persist)
     
     def set_scan_pattern(self, config=None, name=None, persist = False):
         """ Function to set a new scan pattern, see: :any:`protobuf_protocol`.
@@ -313,36 +382,54 @@ class scanner(object):
         req = connection_pb2.Request()
         req.hello.protocol_version = self.protocol_version
         return self._connection.send_request(req).timestamp_ns / 1e9
-    
-    def get_ntp_server(self):
-        """Attention: To use this function you need the requests library
-        `https://requests.readthedocs.io/en/master/`
 
-        Function to get a ntp server
+    def set_time_synchronization(self, ntp=False, ptp=False, persist=True, wait_for_sync=True, max_sync_duration=60):
+        """> Introduced in BSL v2.18 and firmware v1.19
+        
+        Set the type of time synchronization and set basic parameters
 
-        :return: Returns the ntp server IP address
+        Either NTP or PTP has to be set (If both are set NTP will be choosen). It will overwrite the current time synchronization configuration
+
+        :param ntp: Set this field to choose NTPv4 time synchronization, this can be a list of ntp servers or a bool if the standard config should be used
+        :param ptp: Set this field to choose PTP time synchronization, this can be a list of ptp unicast destinations, this will also activate unicast mode and deactivate multicast mode or a bool if the standard config should be used
+        :param persist: Persist time synchronization config on device and reload it after a power-cycle
+        :param wait_for_sync: Wait until device is synchronized. Raises exception if device is not synchronized within max_sync_duration.
+        :param max_sync_duration: Specify maximum time in seconds for synchronization.
+        :return: None.
         """
-        import requests
-        req = requests.get(self.api_path + "network/ntp")
-        req.raise_for_status()
-        return req.json()['data']['server']
-    
-    def set_ntp_server(self, server):
-        """Attention: To use this function you need the requests library
-        `https://requests.readthedocs.io/en/master/`
+        cfg = self.get_advanced_config()
 
-        Function to set a ntp server
+        if ntp and ptp:
+            raise AttributeError("Either provide ntp or ptp configuration. Not both.")
+        if not ntp and not ptp:
+            raise AttributeError("Neither ntp nor ptp is provided.")
 
-        :param server: Server IP to set the ntp server to
-        :type server: str
-        """
-        import requests
-        import json
-        cur_server = self.get_ntp_server()
-        if cur_server != server:
-            req = requests.put(self.api_path + "network/ntp/server", data=json.dumps({ "data": server }), headers={"Content-Type": "application/json"})
-            req.raise_for_status()
-            log.info("Configured NTP server to '%s'. Please power cycle the device to apply changes." % (self.get_ntp_server()))
+        if ntp:
+            if type(ntp) == bool:
+                cfg.time_synchronization.ntp.SetInParent()
+            elif type(ntp) == list:
+                cfg.time_synchronization.ntp.servers[:] = ntp
+            else:
+                cfg.time_synchronization.ntp.CopyFrom(ntp)
+        if ptp:
+            if type(ptp) == bool:
+                cfg.time_synchronization.ptp.SetInParent()
+            elif type(ptp) == list:
+                cfg.time_synchronization.ptp.unicast_destinations[:] = ptp
+            else:
+                cfg.time_synchronization.ptp.CopyFrom(ptp)
+
+        ret = self.set_advanced_config(cfg, persist=persist)
+
+        if wait_for_sync:
+            while max_sync_duration > 0:
+                status = self.get_status()
+                if status.time_synchronization.state == time_synchronization_pb2.TimeSynchronization.State.SYNCED:
+                    return
+                time.sleep(1)
+                max_sync_duration = max_sync_duration - 1
+        
+            raise Exception("Device failed to synchronize. Current state is %s." % (time_synchronization_pb2.TimeSynchronization.State.Name(status.time_synchronization.state)))
             
     def run_self_test(self):
         """> Introduced in BSL v2.10 and firmware v1.9
@@ -361,7 +448,7 @@ class scanner(object):
         
         :return: Newly created :py:class:`blickfeld_scanner.scanner.connection`
         """
-        return connection(self.hostname_or_ip, self.port, self._key_and_cert)
+        return connection(self.hostname_or_ip, self.port, timeout=self.timeout, key_and_cert=self._key_and_cert)
             
     def attempt_error_recovery(self):
         """> Introduced in BSL v2.13 and firmware v1.13
@@ -411,6 +498,15 @@ class scanner(object):
         req = connection_pb2.Request()
         req.delete_named_scan_pattern.name = name
         return self._connection.send_request(req).delete_named_scan_pattern
+
+    def get_imu_stream(self, as_numpy=False):
+        """> Introduced in BSL v2.18 and firmware v1.19
+        
+        Request IMU stream of device
+
+        :returns: :py:class:`blickfeld_scanner.scanner.stream.imu` object
+        """
+        return stream.imu(self.create_connection(), as_numpy=as_numpy)
  
     @staticmethod
     def sync(devices, scan_pattern = None, target_frame_rate = None, max_time_difference = 0.1):
@@ -510,12 +606,14 @@ class connection(object):
     :type hostname_or_ip: str
     :param port: port on which the connection should be established
     :type port: int
+    :param timeout: Timeout in seconds until connection is aborted
+    :type timeout: int
     :param key_and_cert: String containing a private key and certificate for SSL connection
     :type key_and_cert: str
     :param ssl_protocol: The protocol used for an SSL connection
     :type ssl_protocol: see ssl lib: `ssl.PROTOCOL_...`
     """
-    def __init__(self, hostname_or_ip, port, key_and_cert=None, ssl_protocol=ssl.PROTOCOL_TLS):
+    def __init__(self, hostname_or_ip, port, timeout, key_and_cert=None, ssl_protocol=ssl.PROTOCOL_TLS):
         self._key_and_cert = key_and_cert
         self.has_ssl = self._key_and_cert != None
         if self.has_ssl:
@@ -533,11 +631,10 @@ class connection(object):
                 raise
             os.unlink(key_and_cert_file)
             
-            self._ssl_context.verify_mode = ssl.CERT_REQUIRED
-            self.socket = self._ssl_context.wrap_socket(socket.create_connection((hostname_or_ip, port)), server_hostname=hostname_or_ip)
+            self._ssl_context.verify_mode = ssl.CERT_NONE
+            self.socket = self._ssl_context.wrap_socket(socket.create_connection((hostname_or_ip, port), timeout=timeout), server_hostname=hostname_or_ip)
         else:
-            self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
-            self.socket.connect((hostname_or_ip, port))
+            self.socket = socket.create_connection((hostname_or_ip, port), timeout=timeout)
          
     def __del__(self):
         self.close()
@@ -597,9 +694,16 @@ class connection(object):
     def __recv_all(self, n):
         data = b''
         while len(data) < n:
-            packet = self.socket.recv(n - len(data))
+            try:
+                packet = self.socket.recv(n - len(data))
+            except socket.timeout:
+                err = error_pb2.Error()
+                err.connection_abort.details = "Timeout after %u seconds" % (self.socket.timeout)
+                raise protocol_exception(err)
             if not packet:
-                return None
+                err = error_pb2.Error()
+                err.connection_abort.details = "Receive failed"
+                raise protocol_exception(err)
             data += packet
         return data
 
